#include <stdio.h>
#include <windows.h>
#include <assert.h>
#include <sys/stat.h>
#include <io.h>
#include <fcntl.h>
#include <psapi.h>
#include <tlhelp32.h>
#include <signal.h>

#include <map>

static DWORD a_rpisplu;
static DWORD a_kprbprs;
static DWORD a_opiprs;
static DWORD a_OCIStmtPrepare;

static DWORD PID;

static DWORD process_id;
static HANDLE process_handle;

static std::map <DWORD, HANDLE> threads;
static std::map <DWORD, DWORD> TLBs;

static char ORACLE_HOME[1024];

static int oraversion;

static bool detach=false;

void L (const char * fmt, ...)
{
  FILE *f;
  SYSTEMTIME t;
  va_list va;

  GetLocalTime (&t);

  f=fopen ("ospy.log", "a");

  va_start (va, fmt);

  fprintf (f, "[%04d-%02d-%02d %02d:%02d:%02d:%03d] ", t.wYear, t.wMonth, t.wDay, t.wHour, t.wMinute, t.wSecond, t.wMilliseconds);

  vfprintf (f, fmt, va);
  vprintf (fmt, va);

  assert (fclose (f)==0);
};

bool EnableDebugPrivilege( bool Enable ) 
{
  HANDLE hToken = NULL;

  assert (OpenProcessToken( GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, &hToken)!=0);

  TOKEN_PRIVILEGES tp; 

  tp.PrivilegeCount = 1;

  assert (LookupPrivilegeValue( NULL, SE_DEBUG_NAME, &tp.Privileges[0].Luid)!=0);

  tp.Privileges[0].Attributes = Enable ? SE_PRIVILEGE_ENABLED : 0;

  assert (AdjustTokenPrivileges( hToken, FALSE, &tp, sizeof(tp), NULL, NULL)!=0);

  if( hToken != NULL ) 
    assert (CloseHandle( hToken )!=0);

  return true;
}

bool file_exist (char * fname)
{
  int f=_open (fname, _O_RDONLY | _O_BINARY, 0);

  if (f==-1) return false;

  _close (f);

  return true;
};

bool get_oracle_version ()
{
  char * fname=(char*)malloc(10240);
  char * found;
  BYTE * a;
  HMODULE h;
  BYTE v1, v2;

  _snprintf (fname, 10240, "%s\\BIN\\oracommon9.dll", ORACLE_HOME);
  if (file_exist (fname))
    found=fname;
  else
    {
      _snprintf (fname, 10240, "%s\\BIN\\oracommon10.dll", ORACLE_HOME);
      if (file_exist (fname))
	found=fname;
      else
	{
	  _snprintf (fname, 10240, "%s\\BIN\\oracommon11.dll", ORACLE_HOME);
	  if (file_exist (fname))
	    found=fname;
	  else
	    {
	      printf ("Unable to find any of oracommon(9|10|11).dll\n");
	      return false;
	    };
	};
    };

  h=LoadLibrary (found);
  if (h==NULL)
    {
      printf ("Cannot load %s...\n", found);
      return false;
    };
  a=(BYTE*)GetProcAddress (h, "vsnnum");
  if (a==NULL)
    {
      printf ("vsnnum is not found in %s...\n", found);
      return false;
    };
  
  v2=*(a+2);
  v1=*(a+3);

  oraversion=0;

  if (v1==11 && v2==0x10)
    oraversion=111;

  if (v1==10 && v2==0x20)
    oraversion=102;

  if (v1==10 && v2==0x10)
    oraversion=101;
  
  if (v1==9 && v2==0x20)
    oraversion=92;
  
  assert (FreeLibrary (h)!=0);

  free (fname);

  if (oraversion==0)
    {
      printf ("This Oracle version is not supported\n");
      return false;
    }
  else
    return true;
};

bool search_for_symbol_in_SYM (char * sym_fname, char * sym, DWORD * a)
{
  int f=_open (sym_fname, _O_RDONLY | _O_BINARY, 0);
  int len;
  int sig;
  int cnt;
  int something;
  int remain;
  int before;

  DWORD	*d1;
  DWORD	*d2;
  char *d3;

  if (f==-1)
    {
      printf ("Cannot find or open %s\n", sym_fname);
      return false;
    };

  assert ((len=_lseek (f, 0, SEEK_END))!=-1);
  assert (_lseek (f, 0, SEEK_SET)!=-1);

  assert (_read (f, &sig, 4)==4);

  if (sig!=0x4D59534F)
    {
      _close (f);
      printf ("%s is not ORACLE.SYM file (signature mismatch)\n", sym_fname);
      return false;
    };

  assert (_read (f, &cnt, 4)==4);

  before=_tell (f);
  assert (before!=-1L);

  assert (_read (f, &something, 4)==4); // for 11g only

  if (oraversion!=111)
    assert (_lseek (f, before, SEEK_SET)!=-1L);
  else
    len-=4;

  assert (d1=(DWORD*)malloc (cnt*4));
  assert (_read (f, d1, cnt*4)==cnt*4);

  assert (d2=(DWORD*)malloc (cnt*4));
  assert (_read (f, d2, cnt*4)==cnt*4);

  remain=len-8-(cnt*8);
  assert (d3=(char*)malloc (remain));
  assert (_read (f, d3, remain)==remain);

  *a=0;

  for (int i=0;i<cnt;i++)
    {
      if (stricmp (sym, d3+(*(d2+i)))==0)
	{
	  *a=*(d1+i);
	  break;
	};
    };

  free (d1);
  free (d2);
  free (d3);

  _close (f);

  if (*a==0)
    {
      printf ("Symbol %s is not found in %s\n", sym, sym_fname);
      return false;
    }
  else
    return true;
};

bool find_process_by_name (char *name, DWORD * PID)
{
  HANDLE hSysSnapshot=NULL;
  PROCESSENTRY32 proc;
  DWORD rt=0;

  proc.dwSize = sizeof(proc);
  hSysSnapshot = CreateToolhelp32Snapshot ( TH32CS_SNAPPROCESS, 0 );
  if ( hSysSnapshot == (HANDLE)-1 )
    return 0;
  if ( Process32First ( hSysSnapshot, &proc ) )
    {
      proc.dwSize = sizeof(proc);
      do
	{
	  if (stricmp (name, proc.szExeFile)==0 && rt!=0)
	    {
	      printf ("More then one loaded process have name %s.\n", name);
	      return false;
	    };
	  if (stricmp (name, proc.szExeFile)==0 && rt==0)
	    rt=proc.th32ProcessID;
	}
      while ( Process32Next ( hSysSnapshot, &proc ) );
    }
  CloseHandle ( hSysSnapshot );
  *PID=rt;
  return true;
};

DWORD mem_get_dword (HANDLE hProcess, DWORD a)
{
  DWORD val;

  if (ReadProcessMemory (hProcess, (LPCVOID)a, &val, 4, NULL)==FALSE)
    {
      printf ("mem_get_dword(%08X): cannot read memory\n", a);
      assert (0);
    };

  return val;
};

void new_thread (DWORD id, HANDLE h, DWORD TLB)
{
  CONTEXT ctx;

  threads[id]=h;
  TLBs[id]=TLB;

  if (detach==false)
    {
      ctx.ContextFlags = CONTEXT_ALL | CONTEXT_DEBUG_REGISTERS;

      assert (GetThreadContext (h, &ctx)!=0);
 
      ctx.Dr0=a_rpisplu;
      ctx.Dr1=a_kprbprs;
      ctx.Dr2=a_opiprs;
      ctx.Dr3=a_OCIStmtPrepare;
      ctx.Dr7|=(1 | 4 | 0x10 | 0x40);

      assert (SetThreadContext (h, &ctx)!=0);
    };
};

char * get_oracle_process_name (DWORD thread_id)
{
  DWORD a, ThreadLocalStoragePointer;
  DWORD TEB_address;
  char * buf;

  buf=(char*)malloc (1024);

  if (oraversion==111)
    {
      assert (TLBs.find (thread_id)!=TLBs.end());

      TEB_address=TLBs[thread_id];

      a=mem_get_dword (process_handle, TEB_address+0x18);
      ThreadLocalStoragePointer=mem_get_dword (process_handle, TEB_address+0x2C);

      if (ThreadLocalStoragePointer!=0)
	{
	  DWORD ptr=mem_get_dword (process_handle, ThreadLocalStoragePointer);
	  {
	    DWORD v=mem_get_dword (process_handle, ptr+0xF8A4); // 11g win32: ksupga_
	    if (v!=0)
	      if (ReadProcessMemory (process_handle, (LPCVOID)(v+0x490), buf, 1024, NULL)==TRUE)
		return buf;
	  };
	};
    };

  sprintf (buf, "Win32 TID: %d", thread_id);

  return buf;
};

void cleanup_all_threads ()
{
  for (std::map <DWORD, HANDLE>::iterator i=threads.begin(); i!=threads.end(); i++)
    {
      HANDLE h=i->second;

      CONTEXT ctx;

      ctx.ContextFlags = CONTEXT_ALL | CONTEXT_DEBUG_REGISTERS;

      if (GetThreadContext (h, &ctx)!=0)
	{
	  ctx.Dr7=0;
	  SetThreadContext (h, &ctx);
	}
      else
	{
	};
    };
};

void cycle()
{
  DEBUG_EVENT de;
  bool exception_handled;

  while (WaitForDebugEvent (&de, (detach==true) ? 1000 : INFINITE))
    {
      exception_handled=false;

      assert (de.dwThreadId<=0xFFFF);

      switch (de.dwDebugEventCode)
	{
	case EXCEPTION_DEBUG_EVENT:

	  switch (de.u.Exception.ExceptionRecord.ExceptionCode)
	    {
	    case EXCEPTION_SINGLE_STEP:
	      {
		CONTEXT ctx;
		DWORD arg5;
#define STMT_BUF 102400
		BYTE buf[STMT_BUF];

		ctx.ContextFlags = CONTEXT_ALL | CONTEXT_DEBUG_REGISTERS;

		assert (GetThreadContext (threads[de.dwThreadId], &ctx)!=0);

		switch (ctx.Dr6 & 0xF)
		  {
		  case 1:
		    if (ctx.Eip==a_rpisplu)
		      {
			ctx.Dr0=mem_get_dword (process_handle, ctx.Esp); // ret point

			DWORD a;
			if (oraversion==111)
			  a=mem_get_dword (process_handle, ctx.Esp+(5*4)); // 5th arg from 1st
			else
			  a=mem_get_dword (process_handle, ctx.Esp+(4*4)); // 4th arg from 1st

			if (ReadProcessMemory (process_handle, (LPCVOID)a, buf, STMT_BUF, NULL)==TRUE)
			  {
			    char * p_n=get_oracle_process_name (de.dwThreadId);
			    L ("[%s] rpisplu:\n========================================\n%s\n========================================\n\n", p_n, buf);
			    free (p_n);
			  };
		      }
		    else
		      ctx.Dr0=a_rpisplu;

		    break;

		  case 2: // kprbprs
		    if (ctx.Eip==a_kprbprs)
		      {
			ctx.Dr1=mem_get_dword (process_handle, ctx.Esp); // ret point

			DWORD arg3=mem_get_dword (process_handle, ctx.Esp+(3*4));

			if (ReadProcessMemory (process_handle, (LPCVOID)arg3, buf, STMT_BUF, NULL)==TRUE)
			  {
			    char * p_n=get_oracle_process_name (de.dwThreadId);
			    L ("[%s] kprbprs:\n========================================\n%s\n========================================\n\n", p_n, buf);
			    free (p_n);
			  };
		      } 
		    else
		      ctx.Dr1=a_kprbprs;

		    break;

		  case 4: // opiprs
		    
		    if (ctx.Eip==a_opiprs)
		      {
			ctx.Dr2=mem_get_dword (process_handle, ctx.Esp); // ret point

			DWORD arg2=mem_get_dword (process_handle, ctx.Esp+(2*4));
		    
			if (ReadProcessMemory (process_handle, (LPCVOID)arg2, buf, STMT_BUF, NULL)==TRUE)
			  {
			    char * p_n=get_oracle_process_name (de.dwThreadId);
			    L ("[%s] opiprs:\n========================================\n%s\n========================================\n\n", p_n, buf);
			    free (p_n);
			  };
		      }
		    else
		      ctx.Dr2=a_opiprs;
		      
		    break;

		  case 8: // OCIStmtPrepare
		    if (ctx.Eip==a_OCIStmtPrepare)
		      {
			ctx.Dr3=mem_get_dword (process_handle, ctx.Esp); // ret point

			DWORD arg3=mem_get_dword (process_handle, ctx.Esp+(3*4));
		    
			if (ReadProcessMemory (process_handle, (LPCVOID)arg3, buf, STMT_BUF, NULL)==TRUE)		      
			  {
			    char * p_n=get_oracle_process_name (de.dwThreadId);
			    L ("[%s] OCIStmtPrepare:\n========================================\n%s\n========================================\n\n", p_n, buf);
			    free (p_n);
			  };
		      }
		    else
		      ctx.Dr3=a_OCIStmtPrepare;

		    break;

		  default:
		    assert (0);
		  };
	
		assert (SetThreadContext (threads[de.dwThreadId], &ctx)!=0);

		exception_handled=true;
	      };
	      break;

	    case EXCEPTION_BREAKPOINT:
	      exception_handled=true; // for win2k. 2003, XP not send this exception
	      break;

	    default:

	      break;
	    };
	  break;

	case CREATE_PROCESS_DEBUG_EVENT:

	  process_id=de.dwProcessId;
	  process_handle=de.u.CreateProcessInfo.hProcess;
	  
	  new_thread (de.dwThreadId, de.u.CreateProcessInfo.hThread, (DWORD)de.u.CreateProcessInfo.lpThreadLocalBase);
 
	  break;

	case CREATE_THREAD_DEBUG_EVENT: 

	  new_thread (de.dwThreadId, de.u.CreateThread.hThread, (DWORD)de.u.CreateThread.lpThreadLocalBase);

	  break;
	case EXIT_THREAD_DEBUG_EVENT:
	  {
	    assert (threads.find (de.dwThreadId)!=threads.end());
	    threads.erase (threads.find (de.dwThreadId));
	  };
	  break;

	case EXIT_PROCESS_DEBUG_EVENT:
	  break;
	case LOAD_DLL_DEBUG_EVENT:	
	  break;
	case UNLOAD_DLL_DEBUG_EVENT:
	  break;	
	case OUTPUT_DEBUG_STRING_EVENT:	
	  break;
	default:			
	  break;
	};

      if (de.dwDebugEventCode==EXIT_PROCESS_DEBUG_EVENT)
	detach=true;

      if (exception_handled==false)
	ContinueDebugEvent (de.dwProcessId, de.dwThreadId, DBG_EXCEPTION_NOT_HANDLED);

      if (exception_handled==true)
	ContinueDebugEvent (de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
    };
 
};

void __cdecl signal_handler (int signo)
{
  if (signo==SIGINT)
    {
      printf ("Detaching...\n");
      cleanup_all_threads();
      detach=true;
    };
};

void main ()
{
  BOOL (WINAPI * DebugActiveProcessStop_ptr)(DWORD pid);

  _unlink ("ospy.log");
  
  L ("\nOracle SPY, built at %s %s\n -- Dennis Yurichev <dennis@conus.info> http://blogs.conus.info\n\n", __DATE__, __TIME__);

  {
    DWORD r=GetEnvironmentVariable ("ORACLE_HOME", ORACLE_HOME, 1024);
    if (r==0)
      {
	printf ("ORACLE_HOME is not defined.\n");
	return;
      };
    assert (r<=1024);
  };

  if (get_oracle_version()==false)
    return;

  {
    char sympath[1024];
    _snprintf (sympath, 1024, "%s\\RDBMS\\ADMIN\\oracle.sym", ORACLE_HOME);

    if (search_for_symbol_in_SYM (sympath, "_rpisplu", &a_rpisplu)==false) return;
    if (search_for_symbol_in_SYM (sympath, "_kprbprs", &a_kprbprs)==false) return;
    if (search_for_symbol_in_SYM (sympath, "_opiprs", &a_opiprs)==false) return;
    if (search_for_symbol_in_SYM (sympath, "_OCIStmtPrepare", &a_OCIStmtPrepare)==false) return;

    if (find_process_by_name ("oracle.exe", &PID)==false)
      return;

    EnableDebugPrivilege (true);

    DebugActiveProcessStop_ptr=reinterpret_cast<BOOL (WINAPI *)(DWORD)>(GetProcAddress (LoadLibrary ("kernel32.dll"), "DebugActiveProcessStop"));
    if (DebugActiveProcessStop_ptr==NULL)
      printf ("DebugActiveProcessStop() was not detected in your kernel32.dll: detach will not be possible.\n");
    else
      printf ("DebugActiveProcessStop() was detected in your kernel32.dll: detach is possible.\n");

    assert (signal(SIGINT, &signal_handler)!=SIG_ERR);

    if (DebugActiveProcess (PID)==FALSE)
      {
	printf ("Unable to attach to PID %d\n", PID);
	return;
      };

    cycle();

    if (DebugActiveProcessStop_ptr!=NULL)
      assert ((*DebugActiveProcessStop_ptr) (PID)!=0);
    else
      {
	printf ("kernel32.dll!DebugActiveProcessStop() was not found, so we should kill this process.\n");
	assert (TerminateProcess (process_handle, 0)!=0);
      };
  };
};
